"""
Filename: model_training.py
Author: Diptanu Sarkar, ds9297@rit.edu, Rochester Institute of Technology
Description: TBU
"""


# Importing the required libraries
import click
import yaml
import os
import gc
import ast
import torch
from typing import List
from pathlib import Path
import pandas as pd
from pprint import pprint
from sklearn.metrics import f1_score
from farm.modeling.tokenization import Tokenizer
from farm.data_handler.data_silo import DataSilo
from farm.modeling.language_model import LanguageModel
from farm.modeling.prediction_head import (
    TextClassificationHead,
    TokenClassificationHead,
)
from farm.modeling.adaptive_model import AdaptiveModel
from farm.modeling.optimization import initialize_optimizer
from farm.train import Trainer
from farm.train import EarlyStopping
from farm.utils import set_all_seeds, initialize_device_settings
from mtl_processor import MTLProcessor
from farm.evaluation.metrics import register_metrics
from farm.infer import Inferencer


def custom_f1_score(y_true, y_pred):
    f1_scores = []
    for t, p in zip(y_true, y_pred):
        f1_scores.append(f1_score(t, p, average="macro"))
    return {"F1 macro score": sum(f1_scores) / len(f1_scores), "Total": len(f1_scores)}


def loss_function(
    individual_losses: List[torch.Tensor], global_step=None, batch=None, alpha=1, beta=1
):
    loss = (
        alpha * torch.sum(individual_losses[0]) + beta * torch.sum(individual_losses[1])
    ) / (alpha + beta)
    return loss


def train_and_evaluate(parameters: dict):
    device, n_gpu = initialize_device_settings(use_cuda=True)
    TOKEN_LABELS = ["X", "0", "1"]
    LABEL_LIST = ["normal", "offensive", "hatespeech"]

    # Set-up
    test_result_data = pd.read_csv(parameters["test_file"], delimiter=",")
    test_texts = []
    for idx, text in enumerate(test_result_data.post_tokens.values):
        in_dict = {}
        text = ast.literal_eval(text)
        in_dict["text"] = " ".join(text)
        test_texts.append(in_dict)

    # Training
    for random_seed in parameters["random_seeds"]:
        gc.collect()
        torch.cuda.empty_cache()

        # Set the random seed
        set_all_seeds(seed=random_seed)

        os.system("rm -rf /content/early-stopping-model")

        tokenizer = Tokenizer.load(
            pretrained_model_name_or_path=parameters["tokenizer"],
            do_lower_case=parameters["do_lower_case"],
            add_prefix_space="roberta" in parameters["tokenizer"],  # For roberta only
        )

        processor = MTLProcessor(
            data_dir=".",
            tokenizer=tokenizer,
            max_seq_len=parameters["max_seq_size"],
            train_filename=parameters["train_file"],
            test_filename=parameters["test_file"],
            dev_filename=parameters["dev_file"],
            delimiter=",",
        )

        register_metrics("f1_weighted", custom_f1_score)

        processor.add_task(
            name="document_level_task",
            label_list=LABEL_LIST,
            metric="acc",
            text_column_name="text",
            label_column_name="label",
            task_type="classification",
        )
        processor.add_task(
            name="token_level_task",
            label_list=TOKEN_LABELS,
            metric="f1_weighted",
            text_column_name="text",
            label_column_name="tokens",
            task_type="ner",
        )

        data_silo = DataSilo(processor=processor, batch_size=parameters["batch_size"])

        earlystopping = EarlyStopping(
            metric="loss",
            mode="min",
            save_dir=Path("./early-stopping-model"),
            patience=10,
        )

        language_model = LanguageModel.load(parameters["transformer_model"])

        document_level_task_head = TextClassificationHead(
            num_labels=len(LABEL_LIST), task_name="document_level_task"
        )
        token_level_task_head = TokenClassificationHead(
            num_labels=len(TOKEN_LABELS), task_name="token_level_task"
        )

        model = AdaptiveModel(
            language_model=language_model,
            prediction_heads=[document_level_task_head, token_level_task_head],
            embeds_dropout_prob=parameters["embeds_dropout_prob"],
            lm_output_types=["per_sequence", "per_token"],
            device=device,
            loss_aggregation_fn=loss_function,
        )

        model, optimizer, lr_schedule = initialize_optimizer(
            model=model,
            device=device,
            learning_rate=float(parameters["learning_rate"]),
            n_batches=len(data_silo.loaders["train"]),
            n_epochs=parameters["epochs"],
        )

        trainer = Trainer(
            model=model,
            optimizer=optimizer,
            data_silo=data_silo,
            epochs=parameters["epochs"],
            n_gpu=n_gpu,
            lr_schedule=lr_schedule,
            device=device,
            evaluate_every=parameters["evaluate_every"],
            early_stopping=earlystopping if parameters["early_stopping"] else None,
        )

        model = trainer.train()

        save_dir = Path("trained-models/model-seed-" + str(random_seed))

        model.save(save_dir)
        processor.save(save_dir)

        model = Inferencer.load(save_dir, gpu=True)
        result = model.inference_from_dicts(dicts=test_texts)

        label_predictions_list, tokens_predictions_list = [], []
        for idx, chunk_res in enumerate(result):
            if idx % 2 == 0:
                label_predictions_list += chunk_res["predictions"]
            else:
                tokens_predictions_list += chunk_res["predictions"]

        # Tokens predictions
        tokens_list = []
        for idx, pred_ind_list in enumerate(tokens_predictions_list):
            ind_list = []
            for val_dict in pred_ind_list:
                label_val = val_dict["label"]
                ind_list.append(0 if label_val == "X" else int(label_val))
            tokens_list.append(ind_list)
        test_result_data["seed_token_" + str(random_seed)] = tokens_list

        # Labels predictions
        label_list = []
        for idx, pred_dict in enumerate(label_predictions_list):
            label_list.append(pred_dict["label"])
        test_result_data["seed_post_" + str(random_seed)] = label_list

        print("Completed processing:", "seed-" + str(random_seed))

    post_true_values = test_result_data.post_label.values
    token_true_values = test_result_data.toxic_tokens.values

    post_pred_values = []
    for idx in range(len(post_true_values)):
        res_dict = {"offensive": 0, "normal": 0, "hatespeech": 0}

        for random_seed in parameters["random_seeds"]:
            res_dict[test_result_data["seed_post_" + str(random_seed)].values[idx]] += 1

        res_dict = {
            k: v for k, v in sorted(res_dict.items(), key=lambda item: -item[1])
        }
        post_pred_values.append(list(res_dict)[0])

    print("========== Post-level Results ==========")
    for random_seed in parameters["random_seeds"]:
        print(
            "Model seed-",
            random_seed,
            ":",
            f1_score(
                post_true_values,
                test_result_data["seed_post_" + str(random_seed)].values,
                average="macro",
            ),
        )
    print(
        "Overall post-level (macro):",
        f1_score(post_true_values, post_pred_values, average="macro"),
    )

    print("========== Token-level Results ==========")
    for random_seed in parameters["random_seeds"]:
        print(
            "Model seed-",
            random_seed,
            ":",
            res_customr_f1(
                token_true_values,
                test_result_data["seed_token_" + str(random_seed)].values,
            ),
        )
    print(
        "Overall token-level (macro):",
        res_customr_f1(
            token_true_values,
            majority_vote(test_result_data, parameters["random_seeds"]),
        ),
    )


def res_customr_f1(y_true, y_pred):
    f1_scores = []
    idx = 0
    for t, p in zip(y_true, y_pred):
        try:
            t = ast.literal_eval(t)
            cur = f1_score(t, p, average="macro")
            f1_scores.append(cur)
        except Exception:
            diff = len(t) - len(p)
            p = p + [0] * diff
            cur = f1_score(t, p, average="macro")
            f1_scores.append(cur)
        idx += 1
    return "Mean F1 (macro) score: " + str(sum(f1_scores) / len(f1_scores))


def majority_vote(results_df, random_seed_list):
    pred_list = []
    for idx in range(len(results_df)):
        indv_list = []
        for seed in random_seed_list:
            seed_name = "seed_token_" + str(seed)
            seed_list = results_df[seed_name].values[idx]
            if len(indv_list) == 0:
                for i in range(len(seed_list)):
                    indv_list.append(dict({0: 0, 1: 0}))
            for idx_sl, idv_tokens in enumerate(seed_list):
                indv_list[idx_sl][idv_tokens] += 1
        fresh_list = []
        for token_dict in indv_list:
            token_dict = {
                k: v for k, v in sorted(token_dict.items(), key=lambda item: -item[1])
            }
            fresh_list.append(list(token_dict)[0])
        pred_list.append(fresh_list)
    return pred_list


def validate_yaml(parameters: dict):
    try:
        if "train_file" not in parameters and not os.path.isfile(
            parameters["train_file"]
        ):
            raise Exception("Training file not found.")
        if "test_file" not in parameters and not os.path.isfile(
            parameters["test_file"]
        ):
            raise Exception("Test file not found.")
        if (
            "transformer_model" not in parameters
            and "tokenizer" not in parameters
            and parameters["transformer_model"] != parameters["tokenizer"]
        ):
            raise Exception("Model and tokenizer is not same or undefined.")
        if "learning_rate" not in parameters:
            raise Exception("Learning rate is missing or undefined.")
        if "max_seq_size" not in parameters:
            raise Exception("Maximum sequence length is missing or undefined.")
        if "epochs" not in parameters and parameters["epochs"] < 1:
            raise Exception("Number of training epochs is missing or undefined.")
        if "batch_size" not in parameters:
            parameters["batch_size"] = 16
        if "random_seeds" not in parameters:
            parameters["random_seeds"] = [42]
        if "alpha" not in parameters or "beta" not in parameters:
            parameters["alpha"] = 1
            parameters["beta"] = 1
        if "dev_file" not in parameters:
            parameters["dev_split"] = 0.15
        if "early_stopping" not in parameters:
            parameters["early_stopping"] = False
        if "evaluate_every" not in parameters:
            parameters["evaluate_every"] = 100
        if "embeds_dropout_prob" not in parameters:
            parameters["embeds_dropout_prob"] = 0.1
    except Exception as e:
        print("Error: ", e)


def header_printer(header: str):
    print("=" * 35)
    print(header)
    print("=" * 35)


@click.command()
@click.option(
    "--config_yaml",
    required=True,
    type=str,
    help="Config YAML file for model training.",
)
def input_file(config_yaml: str):
    try:
        with open(config_yaml) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
            parameters = data["parameters"]
            validate_yaml(parameters)
            header_printer("The parameters for model training")
            pprint(parameters)
            train_and_evaluate(parameters)
    except Exception as e:
        print("Some error occured! Error:", e)


if __name__ == "__main__":
    input_file()



"""
Filename: language_modeler.py
Author: Diptanu Sarkar, ds9297@rit.edu, Rochester Institute of Technology
Description: TBU
"""


# Importing the required libraries
from pprint import pprint
from transformers import (
    LineByLineTextDataset,
    DataCollatorForLanguageModeling,
    Trainer,
    TrainingArguments,
    AutoTokenizer,
    AutoModelForMaskedLM,
)
import click
import yaml
import os


# Training batch size
PER_DEVICE_TRAIN_BATCH_SIZE = 16


def training_model(parameters: dict):
    tokenizer = AutoTokenizer.from_pretrained(parameters["tokenizer"])
    model = AutoModelForMaskedLM.from_pretrained(parameters["transformer_model"])
    dataset = LineByLineTextDataset(
        tokenizer=tokenizer,
        file_path=parameters["retrain_text_filepath"],
        block_size=parameters["max_seq_size"],
    )
    data_collator = DataCollatorForLanguageModeling(
        tokenizer=tokenizer, mlm=True, mlm_probability=parameters["mlm_probability"]
    )
    training_args = TrainingArguments(
        output_dir=parameters["interim_dir"],
        overwrite_output_dir=True,
        num_train_epochs=parameters["epochs"],
        per_device_train_batch_size=PER_DEVICE_TRAIN_BATCH_SIZE,
        save_steps=parameters["save_every"],
        save_total_limit=parameters["save_limit"],
        seed=parameters["random_seed"],
    )
    trainer = Trainer(
        model=model,
        args=training_args,
        data_collator=data_collator,
        train_dataset=dataset,
    )
    header_printer("Starting the training.")
    trainer.train()
    trainer.save_model(parameters["model_save_path"])
    tokenizer.save_pretrained(parameters["model_save_path"])
    header_printer("Training completed.")


def validate_yaml(parameters: dict):
    try:
        if "retrain_text_filepath" not in parameters and not os.path.isfile(
            parameters["retrain_text_filepath"]
        ):
            raise Exception("retraining text file not found.")
        if (
            "transformer_model" not in parameters
            and "tokenizer" not in parameters
            and parameters["transformer_model"] != parameters["tokenizer"]
        ):
            raise Exception("Model and tokenizer is not same or undefined.")
        if "epochs" not in parameters and parameters["epochs"] < 1:
            raise Exception("Number of training epochs is missing or undefined.")
        if "max_seq_size" not in parameters:
            parameters["max_seq_size"] = 128
        if "mlm_probability" not in parameters:
            parameters["mlm_probability"] = 0.15
        if "random_seed" not in parameters:
            parameters["random_seed"] = 42
        if "interim_dir" not in parameters:
            parameters["interim_dir"] = "./retrain-interim"
        if "model_save_path" not in parameters:
            parameters["model_save_path"] = "./custom-model"
        if "save_every" not in parameters:
            parameters["save_every"] = 500
        if "save_limit" not in parameters:
            parameters["save_limit"] = 2
    except Exception as e:
        print("Error: ", e)


def header_printer(header: str):
    print("=" * 30)
    print(header)
    print("=" * 30)


@click.command()
@click.option(
    "--config_yaml",
    required=True,
    type=str,
    help="Config YAML file for language modeling.",
)
def input_file(config_yaml: str):
    try:
        with open(config_yaml) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
            parameters = data["parameters"]
            validate_yaml(parameters)
            header_printer("The parameters for retraining")
            pprint(parameters)
            training_model(parameters)
    except Exception:
        print("Try again! Improper file format or does not exist.")


if __name__ == "__main__":
    input_file()



import os


class MADApp:

    def __init__(self):
        print("INFO: Initializing MAD Framework")
        print("INFO: Installing required packages and models.")
        self.install_farm()
        self.pretrained_model()
        
    
    def install_farm(self):
        try:
            os.system("pip install git+https://github.com/imdiptanu/FARM.git --quiet")
        except Exception as e:
            print("ERROR: Installing FARM! Error:", e)


    def pretrained_model(self):
        try:
            print("INFO: Downloading mad-bert-uncased model from remote.")
            os.system("wget https://www.dropbox.com/s/sfvwyn3bw49x93j/mad-bert-uncased.zip")
            cur_path = os.getcwd()
            model_path = cur_path + "/mad-bert-uncased.zip"
            os.system(f"unzip {model_path}")
            print("INFO: Model downloaded Successfully.")
        except Exception as e:
            print("ERROR: Downloading pre-trained model! Error:", e)


    def mad_print(self, text):
        print(text)



from farm.data_handler.processor import Processor
from tokenizers.pre_tokenizers import WhitespaceSplit
from farm.data_handler.samples import (
    Sample,
    SampleBasket,
)
from farm.data_handler.utils import expand_labels
import ast
import numpy as np
import pandas as pd


class MTLProcessor(Processor):
    def __init__(
        self,
        tokenizer,
        max_seq_len,
        data_dir,
        train_filename,
        test_filename,
        delimiter,
        dev_split=0.15,
        dev_filename=None,
        label_list=None,
        metric=None,
        proxies=None,
        **kwargs,
    ):
        self.delimiter = delimiter

        super(MTLProcessor, self).__init__(
            tokenizer=tokenizer,
            max_seq_len=max_seq_len,
            train_filename=train_filename,
            dev_filename=dev_filename,
            test_filename=test_filename,
            dev_split=dev_split,
            data_dir=data_dir,
            tasks={},
            proxies=proxies,
        )

    def file_to_dicts(self, file: str):
        dicts = list()
        df = pd.read_csv(file)
        for text, label, tokens in zip(
            df.post_tokens.values, df.post_label.values, df.toxic_tokens.values
        ):
            columns = dict()
            text = ast.literal_eval(text)
            tokens = ast.literal_eval(tokens)
            columns["text"] = " ".join(text)
            columns["document_level_task_label"] = label  # Key hard-coded
            columns["token_level_task_label"] = list(map(str, tokens))  # Key hard-coded
            dicts.append(columns)
        return dicts

    @staticmethod
    def _get_start_of_word(word_ids):
        words = np.array(word_ids)
        words[words == None] = -1
        start_of_word_single = [0] + list(np.ediff1d(words) > 0)
        start_of_word_single = [int(x) for x in start_of_word_single]
        return start_of_word_single

    # Most of the code is copied from NERProcessor - dataset_from_dicts()
    def dataset_from_dicts(
        self, dicts, indices=None, return_baskets=False, non_initial_token="X"
    ):
        self.baskets = []
        self.pre_tokenizer = WhitespaceSplit()

        texts = [x["text"] for x in dicts]
        words_and_spans = [self.pre_tokenizer.pre_tokenize_str(x) for x in texts]
        words = [[x[0] for x in y] for y in words_and_spans]

        word_spans_batch = [[x[1] for x in y] for y in words_and_spans]

        tokenized_batch = self.tokenizer.batch_encode_plus(
            words,
            return_offsets_mapping=True,
            return_special_tokens_mask=True,
            return_token_type_ids=True,
            return_attention_mask=True,
            truncation=True,
            max_length=self.max_seq_len,
            padding="max_length",
            is_split_into_words=True,
        )

        for i in range(len(dicts)):
            tokenized = tokenized_batch[i]
            d = dicts[i]
            id_external = self._id_from_dict(d)
            if indices:
                id_internal = indices[i]
            else:
                id_internal = i

            input_ids = tokenized.ids
            segment_ids = tokenized.type_ids
            initial_mask = self._get_start_of_word(tokenized.words)
            assert len(initial_mask) == len(input_ids)

            padding_mask = tokenized.attention_mask

            if return_baskets:
                token_to_word_map = tokenized.words
                word_spans = word_spans_batch[i]
                tokenized_dict = {
                    "tokens": tokenized.tokens,
                    "word_spans": word_spans,
                    "token_to_word_map": token_to_word_map,
                    "start_of_word": initial_mask,
                }
            else:
                tokenized_dict = {}

            feature_dict = {
                "input_ids": input_ids,
                "padding_mask": padding_mask,
                "segment_ids": segment_ids,
                "initial_mask": initial_mask,
            }

            for task_name, task in self.tasks.items():
                try:
                    label_name = task["label_name"]
                    labels_word = d[label_name]
                    label_list = task["label_list"]
                    label_tensor_name = task["label_tensor_name"]

                    if task["task_type"] == "classification":
                        label_ids = [label_list.index(labels_word)]
                    elif task["task_type"] == "ner":
                        labels_token = expand_labels(
                            labels_word, initial_mask, non_initial_token
                        )
                        label_ids = [label_list.index(lt) for lt in labels_token]
                except ValueError:
                    label_ids = None
                    problematic_labels = set(labels_token).difference(set(label_list))
                    print(
                        f"[Task: {task_name}] Could not convert labels to ids via label_list!"
                        f"\nWe found a problem with labels {str(problematic_labels)}"
                    )
                except KeyError:
                    label_ids = None
                if label_ids:
                    feature_dict[label_tensor_name] = label_ids

            curr_sample = Sample(
                id=None, clear_text=d, tokenized=tokenized_dict, features=[feature_dict]
            )
            curr_basket = SampleBasket(
                id_internal=id_internal,
                raw=d,
                id_external=id_external,
                samples=[curr_sample],
            )
            self.baskets.append(curr_basket)

        if indices and 0 not in indices:
            pass
        else:
            self._log_samples(1)

        dataset, tensor_names = self._create_dataset()
        ret = [dataset, tensor_names, self.problematic_sample_ids]
        if return_baskets:
            ret.append(self.baskets)
        return tuple(ret)



from distutils.core import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mad_fw",
    packages = ["mad_fw"],
    version="0.1.11",
    license="MIT",
    description="Multitask Aggression Detection (MAD)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Diptanu Sarkar",
    author_email="diptanusarkar@hotmail.com",
    url="https://github.com/imdiptanu/MAD",
    download_url="https://github.com/imdiptanu/MAD/archive/refs/tags/0.1.4.tar.gz",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.7",
    install_requires=[
        "click",
        "pathlib",
        "pandas",
        "scikit-learn",
        "pyyaml",
        "typing",
    ],
)
